@@ -65,7 +65,7 @@ class ReversibleSelfAttnBlock(nn.Module):
         self.j = Deterministic(j)
         self.k = Deterministic(k)        
 
-    def forward(self, x, m, mask = None, msa_mask = None, _reverse = True):
+    def forward(self, x, m, mask = None, msa_mask = None, seq_shape = None, msa_shape = None, _reverse = True):
         x1, x2 = torch.chunk(x, 2, dim = 2)
         m1, m2 = torch.chunk(m, 2, dim = 2)
         y1, y2, n1, n2 = None, None, None, None
@@ -74,14 +74,15 @@ class ReversibleSelfAttnBlock(nn.Module):
         record_rng = self.training and _reverse
 
         with context():
-            y1 = x1 + self.f(x2, record_rng = record_rng, mask = mask)
+            print(x2.shape, seq_shape)
+            y1 = x1 + self.f(x2, seq_shape, record_rng = record_rng, mask = mask, )
             y2 = x2 + self.g(y1, record_rng = record_rng)
-            n1 = m1 + self.j(m2, record_rng = record_rng, mask = msa_mask)
+            n1 = m1 + self.j(m2, msa_shape, record_rng = record_rng, mask = msa_mask)
             n2 = m2 + self.k(n1, record_rng = record_rng)
 
         return torch.cat((y1, y2), dim = 2), torch.cat((n1, n2), dim = 2)
 
-    def backward_pass(self, y, n, dy, dn, mask = None, msa_mask = None):
+    def backward_pass(self, y, n, dy, dn, mask = None, msa_mask = None, seq_shape = None, msa_shape = None):
         y1, y2 = torch.chunk(y, 2, dim = 2)
         del y
 
@@ -103,7 +104,7 @@ class ReversibleSelfAttnBlock(nn.Module):
 
         with torch.enable_grad():
             x2.requires_grad = True
-            fx2 = self.f(x2, set_rng = True, mask = mask)
+            fx2 = self.f(x2, seq_shape, set_rng = True, mask = mask)
             torch.autograd.backward(fx2, dx1, retain_graph = True)
 
         with torch.no_grad():
@@ -138,7 +139,7 @@ class ReversibleSelfAttnBlock(nn.Module):
 
         with torch.enable_grad():
             m2.requires_grad = True
-            fm2 = self.j(m2, set_rng = True, mask = msa_mask)
+            fm2 = self.j(m2, msa_shape, set_rng = True, mask = msa_mask)
             torch.autograd.backward(fm2, dm1, retain_graph=True)
 
         with torch.no_grad():
@@ -164,7 +165,7 @@ class ReversibleCrossAttnBlock(nn.Module):
         self.j = Deterministic(j)
         self.k = Deterministic(k)
 
-    def forward(self, x, m, mask = None, msa_mask = None, _reverse = True):
+    def forward(self, x, m, mask = None, msa_mask = None, _reverse = True, **kwargs):
         x1, x2 = torch.chunk(x, 2, dim = 2)
         m1, m2 = torch.chunk(m, 2, dim = 2)
         y1, y2, n1, n2 = None, None, None, None
@@ -180,7 +181,7 @@ class ReversibleCrossAttnBlock(nn.Module):
 
         return torch.cat((y1, y2), dim = 2), torch.cat((n1, n2), dim = 2)
 
-    def backward_pass(self, y, n, dy, dn, mask = None, msa_mask = None):
+    def backward_pass(self, y, n, dy, dn, mask = None, msa_mask = None, **kwargs):
         n1, n2 = torch.chunk(n, 2, dim = 2)
         del n
 
@@ -311,12 +312,12 @@ class ReversibleSequence(nn.Module):
 
         self.blocks = blocks
 
-    def forward(self, seq, msa, mask = None, msa_mask = None, reverse = True):
+    def forward(self, seq, msa, seq_shape = None, msa_shape = None, mask = None, msa_mask = None, reverse = True):
         assert exists(msa), 'reversibility does not work with no MSA sequences yet'
         
         blocks = self.blocks
         seq, msa = list(map(lambda t: torch.cat((t, t), dim = -1), (seq, msa)))
-        kwargs = {'mask': mask, 'msa_mask': msa_mask}
+        kwargs = {'mask': mask, 'msa_mask': msa_mask, 'seq_shape': seq_shape, 'msa_shape': msa_shape}
 
         fn = reversible_apply if reverse else irreversible_apply
         ind = seq.shape[1]
