@@ -6,14 +6,14 @@ from functools import wraps
 from einops import rearrange, repeat
 
 # bio
-import mdtraj
+
 try:
-    from sidechainnet.utils.sequence import ProteinVocabulary as VOCAB
+    from sidechainnet.utils.sequence import ProteinVocabulary, ONE_TO_THREE_LETTER_MAP
     from sidechainnet.utils.measure import GLOBAL_PAD_CHAR
     from sidechainnet.structure.build_info import NUM_COORDS_PER_RES, BB_BUILD_INFO, SC_BUILD_INFO
     from sidechainnet.structure.StructureBuilder import _get_residue_build_iter
     # build vocablary
-    VOCAB = VOCAB()
+    VOCAB = ProteinVocabulary()
 except:
     NUM_COORDS_PER_RES = 14
     GLOBAL_PAD_CHAR = 0
@@ -84,6 +84,48 @@ def invoke_torch_or_numpy(torch_fn, numpy_fn):
         return inner
     return outer
 
+
+# preprocess data
+
+def get_atom_ids_dict():
+    """ Get's a dict mapping each atom to a token. """
+    ids = set(["", "N", "CA", "C", "O"])
+
+    for k,v in SC_BUILD_INFO.items():
+        for name in v["atom-names"]:
+            ids.add(name)
+            
+    return {k: i for i,k in enumerate(sorted(ids))}
+
+def make_cloud_mask(aa):
+    """ relevent points will be 1. paddings will be 0. """
+    mask = np.zeros(14)
+    # early stop if padding token
+    if aa == "_":
+        return mask
+    # get num of atoms in aa
+    n_atoms = 4+len( SC_BUILD_INFO[ ONE_TO_THREE_LETTER_MAP[aa] ]["atom-names"] )
+    mask[:n_atoms] = 1
+    return mask
+
+def make_atom_id_embedds(aa, atom_ids):
+    """ Return the tokens for each atom in the aa. """
+    mask = np.zeros(14)
+    # early stop if padding token
+    if aa == "_":
+        return mask
+    # get atom id
+    atom_list = ["N", "CA", "C", "O"] + SC_BUILD_INFO[ ONE_TO_THREE_LETTER_MAP[aa] ]["atom-names"]
+    for i,atom in enumerate(atom_list):
+        mask[i] = ATOM_IDS[atom]
+    return mask
+
+
+ATOM_IDS = get_atom_ids_dict()
+CUSTOM_INFO = {k: {"cloud_mask": make_cloud_mask(k),
+                   "atom_id_embedd": make_atom_id_embedds(k, atom_ids=ATOM_IDS),
+                  } for k in "ARNDCQEGHILKMFPSTWYV_"}
+
 # common utils
 
 # parsing to pdb for easier visualization - other example from sidechainnet is:
@@ -107,6 +149,7 @@ def clean_pdb(name, route=None, chain_num=None):
         * chain_num: int. index of chain to select (1-indexed as pdb files)
         Output: route of destin file.
     """
+    import mdtraj
     destin = route if route is not None else name
     # read input
     raw_prot = mdtraj.load_pdb(name)
@@ -138,6 +181,7 @@ def custom2pdb(coords, proteinnet_id, route):
         * route: str. destin route.
         Output: tuple of routes: (original, generated) for the structures. 
     """
+    import mdtraj
     # convert to numpy
     if isinstance(coords, torch.Tensor):
         coords = coords.detach().cpu().numpy()
@@ -167,35 +211,56 @@ def scn_cloud_mask(scn_seq, boolean=True):
         * boolean: whether to return as array of idxs or boolean values
         Outputs: (batch, length, NUM_COORDS_PER_RES) boolean mask 
     """
-    # scaffolds 
-    mask = torch.zeros(*scn_seq.shape, NUM_COORDS_PER_RES, device=scn_seq.device)
-    # fill 
-    for n, seq in enumerate(scn_seq.cpu().numpy()):
-        for i,aa in enumerate(seq):
-            # get num of atom positions - backbone is 4: ...N-C-C(=O)...
-            n_atoms = 4+len( SC_BUILD_INFO[VOCAB.int2chars(aa)]["atom-names"] )
-            mask[n, i, :n_atoms] = 1
+    device = scn_seq.device
+    batch_mask = []
+    # do loop in cpu
+    scn_seq = scn_seq.cpu()
+    for i, seq in enumerate(scn_seq):
+        # get masks for each prot (points for each aa)
+        batch_mask.append( torch.tensor([CUSTOM_INFO[VOCAB(aa)]['cloud_mask'] \
+                                         for aa in seq]).bool().to(device).unsqueeze(0) )
+    # concat in last dim
+    batch_mask = torch.cat(batch_mask, dim=0)
     if boolean:
         return mask.bool()
     return mask.nonzero()
 
-def scn_backbone_mask(scn_seq, boolean=True, l_aa=NUM_COORDS_PER_RES):
+def scn_backbone_mask(scn_seq, boolean=True):
     """ Gets the boolean mask for N and CA positions. 
         Inputs: 
-        * scn_seq: sequence as provided by Sidechainnet package
+        * scn_seq: sequence(s) as provided by Sidechainnet package (int tensor/s)
         * bool: whether to return as array of idxs or boolean values
-        Outputs: (N_mask, CA_mask)
+        Outputs: (N_mask, CA_mask, C_mask)
     """
-    lengths = torch.arange(scn_seq.shape[-1]*l_aa)
-    # repeat if needed:
-    if len(lengths.shape) == 2:
-        lengths = repeat(lengths, 'l -> b l', b=scn_seq.shape[0])
+    wrapper = torch.zeros(*scn_seq.shape, 14)
     # N is the first atom in every AA. CA is the 2nd.
-    N_mask  = lengths%l_aa == 0
-    CA_mask = lengths%l_aa == 1
+    wrapper[:, 0] = 1
+    wrapper[:, 1] = 2
+    wrapper[:, 2] = 3
+    wrapper = rearrange(wrapper, '... l c -> ... (l c)')
+    # find idxs
+    N_mask  = wrapper == 1
+    CA_mask = wrapper == 2
+    C_mask  = wrapper == 3 
     if boolean:
-        return N_mask, CA_mask
-    return N_mask.nonzero(), CA_mask.nonzero()
+        return N_mask, CA_mask, C_mask
+    return N_mask.nonzero(), CA_mask.nonzero(), C_mask.nonzero()
+
+def scn_atom_embedd(scn_seq):
+    """ Returns the token for each atom in the aa. 
+        Inputs: 
+        * scn_seq: sequence(s) as provided by Sidechainnet package (int tensor/s)
+    """
+    device = scn_seq.device
+    batch_tokens = []
+    # do loop in cpu
+    scn_seq = scn_seq.cpu()
+    for i,seq in enumerate(scn_seq):
+        batch_tokens.append( torch.tensor([CUSTOM_INFO[k]["atom_id_embedd"] \
+                                           for k in seq]).long().to(device).unsqueeze(0) )
+    batch_tokens = torch.cat(batch_tokens, dim=0)
+    return batch_tokens
+
 
 def nerf_torch(a, b, c, l, theta, chi):
     """ Custom Natural extension of Reference Frame. 
@@ -225,25 +290,37 @@ def nerf_torch(a, b, c, l, theta, chi):
     # extend base point, set length
     return c + l.unsqueeze(-1) * torch.matmul(rotate, d).squeeze()
 
-def sidechain_container(backbones, place_oxygen=False,
+def sidechain_container(backbones, n_aa, cloud_mask=None, place_oxygen=False,
                         n_atoms=NUM_COORDS_PER_RES, padding=GLOBAL_PAD_CHAR):
     """ Gets a backbone of the protein, returns the whole coordinates
         with sidechains (same format as sidechainnet). Keeps differentiability.
         Inputs: 
         * backbones: (batch, L*3, 3): assume batch=1 (could be extended later).
                     Coords for (N-term, C-alpha, C-term) of every aa.
+        * n_aa: int. number of points for each aa in the backbones.
+        * cloud_mask: (batch, l, c). optional. cloud mask from scn_cloud_mask`.
+                      returns point outside to 0. if passed, else c_alpha
         * place_oxygen: whether to claculate the oxygen of the
                         carbonyl group via NeRF
         * n_atoms: int. n of atom positions / atom. same as in sidechainnet: 14
         * padding: int. padding token. same as in sidechainnet: 0
         Outputs: whole coordinates of shape (batch, L, n_atoms, 3)
     """
-    batch, length = backbones.shape[0], backbones.shape[1] // 3
-    new_coords = torch.ones(batch, length, NUM_COORDS_PER_RES, 3, device=backbones.device) * padding
-    new_coords[:, :, :3] = rearrange(backbones, 'b (l back) d -> b l back d', l=length)
-    # set the rest of positions to c_alpha
-    new_coords[:, :, 3:] = repeat(new_coords[:, :, 2], 'b l d -> b l scn d', scn=11)
-    # hard-calculate oxygen position of carbonyl group
+    device = backbones.device
+    batch, length = backbones.shape[0], backbones.shape[1] // n_aa
+    # build scaffold from (N, CA, C, CB)
+    new_coords = torch.zeros(batch, length, NUM_COORDS_PER_RES, 3).to(device)
+    print(new_coords.shape)
+    predicted  = rearrange(backbones, 'b (l back) d -> b l back d', l=length)
+    # set backbone positions
+    new_coords[:, :, :3] = predicted[:, :, :3]
+    # set rest of positions to c_alpha
+    new_coords[:, :, 3:] = repeat(new_coords[:, :, 1], 'b l d -> b l scn d', scn=11)
+    if cloud_mask is not None:
+        new_coords[torch.logical_not(cloud_mask)] = 0.
+    # overwrite cbeta
+    new_coords[:, :, 4] = predicted[:, :, 3]
+    # hard-calculate oxygen position of carbonyl group with parallel version of NERF
     if place_oxygen: 
         # build (=O) position of revery aa in each chain
         for s in range(batch):
@@ -301,7 +378,6 @@ def center_distogram_torch(distogram, bins=DISTANCE_THRESHOLDS, min_t=1., center
     if wide == "var":
         dispersion = (distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1)
     elif wide == "std":
-
         dispersion = (distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1).sqrt()
     else:
         dispersion = torch.zeros_like(central, device=device)
@@ -314,12 +390,12 @@ def center_distogram_torch(distogram, bins=DISTANCE_THRESHOLDS, min_t=1., center
 
 def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     """ Gets distance matrix. Outputs 3d. See below for wrapper. 
-        Assumes (for now) distrogram is (N x N) and symmetric
+        Assumes (for now) distogram is (N x N) and symmetric
         Outs: 
-        * best_3d_coords: (3 x N)
-        * historic_stress 
+        * best_3d_coords: (batch x 3 x N)
+        * historic_stresses: (batch x steps)
     """
-    device = pre_dist_mat.device
+    device, dtype = pre_dist_mat.device, pre_dist_mat.type()
 
     if weights is None:
         weights = torch.ones_like(pre_dist_mat)
@@ -332,8 +408,8 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     diag_idxs = np.arange(N)
     his = []
     # init random coords
-    best_stress = float("Inf") * torch.ones(batch, device = device)
-    best_3d_coords = 2*torch.rand(batch, N, 3, device = device) - 1
+    best_stress = float("Inf") * torch.ones(batch, device = device).type(dtype)
+    best_3d_coords = 2*torch.rand(batch, N, 3, device = device).type(dtype) - 1
     # iterative updates:
     for i in range(iters):
         # compute distance matrix of coords and stress
@@ -467,7 +543,7 @@ def calc_phis_torch(pred_coords, N_mask, CA_mask, C_mask=None,
     c_alphas = pred_coords_[:, CA_mask.squeeze()]
     # select c_term auto if not passed
     if C_mask is not None: 
-        c_terms = pred_coords_[:, C_mask]
+        c_terms = pred_coords_[:, C_mask.squeeze()]
     else:
         c_terms  = pred_coords_[ :, torch.logical_not(torch.logical_or(N_mask,CA_mask)).squeeze() ]
     # compute phis for every pritein in the batch
@@ -645,12 +721,11 @@ def mdscaling_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5,
 
     # no need to caculate multiple mirrors - just correct Z axis
     phi_ratios = calc_phis_torch(preds, N_mask, CA_mask, C_mask, prop=True)
-    for i,pred in enumerate(preds):
-        # fix mirrors by (-1)*Z if more (+) than (-) phi angles
-        if phi_ratios < 0.5:
-            preds[i, -1] = (-1)*preds[i, -1]
-            if verbose == 2:
-                print("Corrected mirror in struct no.", i)
+    to_correct = torch.nonzero( (phi_ratios < 0.5) ).view(-1)
+    # fix mirrors by (-1)*Z if more (+) than (-) phi angles
+    preds[to_correct, -1] = (-1)*preds[to_correct, -1]
+    if verbose == 2:
+        print("Corrected mirror idxs:", to_correct)
             
     return preds, stresses
 
